@@ -93,14 +93,16 @@ static uint32 read_keypart_length(const uchar *from, uint bytes)
 
 // @param sortlen  [Maximum] length of the sort key
 void Sort_param::init_for_filesort(uint sortlen, TABLE *table,
-                                   ha_rows maxrows, bool sort_positions)
+                                   ha_rows maxrows, Filesort *filesort)
 {
   DBUG_ASSERT(addon_fields == NULL);
 
   sort_length= sortlen;
   ref_length= table->file->ref_length;
+  accepted_rows= filesort->accepted_rows;
+
   if (!(table->file->ha_table_flags() & HA_FAST_KEY_READ) &&
-      !table->fulltext_searched && !sort_positions)
+      !table->fulltext_searched && !filesort->sort_positions)
   {
     /* 
       Get the descriptors of all fields whose values are appended 
@@ -196,16 +198,15 @@ SORT_INFO *filesort(THD *thd, TABLE *table, Filesort *filesort,
   size_t memory_available= (size_t)thd->variables.sortbuff_size;
   uint maxbuffer;
   Merge_chunk *buffpek;
-  ha_rows num_rows= HA_POS_ERROR;
+  ha_rows num_rows= HA_POS_ERROR, not_used=0;
   IO_CACHE tempfile, buffpek_pointers, *outfile; 
   Sort_param param;
   bool allow_packing_for_sortkeys;
   Bounded_queue<uchar, uchar> pq;
   SQL_SELECT *const select= filesort->select;
   ha_rows max_rows= filesort->limit;
-  uint s_length= 0;
+  uint s_length= 0, sort_len;
   Sort_keys *sort_keys;
-
   DBUG_ENTER("filesort");
 
   if (!(sort_keys= filesort->make_sortorder(thd, join, first_table_bit)))
@@ -247,9 +248,13 @@ SORT_INFO *filesort(THD *thd, TABLE *table, Filesort *filesort,
   sort->found_rows= HA_POS_ERROR;
 
   param.sort_keys= sort_keys;
-  uint sort_len= sortlength(thd, sort_keys, &allow_packing_for_sortkeys);
+  sort_len= sortlength(thd, sort_keys, &allow_packing_for_sortkeys);
+  param.init_for_filesort(sort_len, table, max_rows, filesort);
+  if (!param.accepted_rows)
+    param.accepted_rows= &not_used;
 
-  param.init_for_filesort(sort_len, table, max_rows, filesort->sort_positions);
+  param.set_all_read_bits= filesort->set_all_read_bits;
+  param.unpack= filesort->unpack;
 
   sort->addon_fields=  param.addon_fields;
   sort->sort_keys= param.sort_keys;
@@ -477,7 +482,8 @@ SORT_INFO *filesort(THD *thd, TABLE *table, Filesort *filesort,
                     thd->killed == ABORT_QUERY ? "LIMIT ROWS EXAMINED" :
                     thd->get_stmt_da()->message());
 
-    if (global_system_variables.log_warnings > 1)
+    if ((thd->killed == ABORT_QUERY || kill_errno) &&
+        global_system_variables.log_warnings > 1)
     { 
       sql_print_warning("%s, host: %s, user: %s, thread: %lu, query: %-.4096s",
                         ER_THD(thd, ER_FILSORT_ABORT),
@@ -671,7 +677,7 @@ const char* dbug_print_row(TABLE *table, const uchar *rec, bool print_names)
 
   output.length(0);
   output.append(table->alias);
-  output.append("(");
+  output.append('(');
   bool first= true;
   if (print_names)
   {
@@ -683,13 +689,13 @@ const char* dbug_print_row(TABLE *table, const uchar *rec, bool print_names)
       if (first)
         first= false;
       else
-        output.append(", ");
+        output.append(STRING_WITH_LEN(", "));
 
-      output.append((*pfield)->field_name.str ?
-                    (*pfield)->field_name.str: "NULL");
+      output.append((*pfield)->field_name.str
+                    ? (*pfield)->field_name : NULL_clex_str);
     }
 
-    output.append(")=(");
+    output.append(STRING_WITH_LEN(")=("));
     first= true;
   }
 
@@ -703,10 +709,10 @@ const char* dbug_print_row(TABLE *table, const uchar *rec, bool print_names)
     if (first)
       first= false;
     else
-      output.append(", ");
+      output.append(STRING_WITH_LEN(", "));
 
     if (field->is_null())
-      output.append("NULL");
+      output.append(&NULL_clex_str);
     else
     {
       if (field->type() == MYSQL_TYPE_BIT)
@@ -716,8 +722,8 @@ const char* dbug_print_row(TABLE *table, const uchar *rec, bool print_names)
       output.append(tmp.ptr(), tmp.length());
     }
   }
-  output.append(")");
-  
+  output.append(')');
+
   return output.c_ptr_safe();
 }
 
@@ -899,6 +905,8 @@ static ha_rows find_all_keys(THD *thd, Sort_param *param, SQL_SELECT *select,
       goto err;
   }
 
+  if (param->set_all_read_bits)
+    sort_form->column_bitmaps_set(save_read_set, save_write_set);
   DEBUG_SYNC(thd, "after_index_merge_phase1");
 
   for (;;)
@@ -906,7 +914,11 @@ static ha_rows find_all_keys(THD *thd, Sort_param *param, SQL_SELECT *select,
     if (quick_select)
       error= select->quick->get_next();
     else					/* Not quick-select */
+    {
       error= file->ha_rnd_next(sort_form->record[0]);
+      if (param->unpack)
+        param->unpack(sort_form);
+    }
     if (unlikely(error))
       break;
     file->position(sort_form->record[0]);
@@ -975,6 +987,7 @@ static ha_rows find_all_keys(THD *thd, Sort_param *param, SQL_SELECT *select,
         idx++;
       }
       num_records++;
+      (*param->accepted_rows)++;
     }
 
     /* It does not make sense to read more keys in case of a fatal error */
@@ -1114,7 +1127,7 @@ Type_handler_string_result::make_sort_key_part(uchar *to, Item *item,
                                             String *tmp_buffer) const
 {
   CHARSET_INFO *cs= item->collation.collation;
-  bool maybe_null= item->maybe_null;
+  bool maybe_null= item->maybe_null();
 
   if (maybe_null)
     *to++= 1;
@@ -1184,7 +1197,7 @@ Type_handler_int_result::make_sort_key_part(uchar *to, Item *item,
                                             String *tmp_buffer) const
 {
   longlong value= item->val_int_result();
-  make_sort_key_longlong(to, item->maybe_null, item->null_value,
+  make_sort_key_longlong(to, item->maybe_null(), item->null_value,
                          item->unsigned_flag, value);
 }
 
@@ -1200,13 +1213,13 @@ Type_handler_temporal_result::make_sort_key_part(uchar *to, Item *item,
   static const Temporal::Options opt(TIME_INVALID_DATES, TIME_FRAC_NONE);
   if (item->get_date_result(current_thd, &buf, opt))
   {
-    DBUG_ASSERT(item->maybe_null);
+    DBUG_ASSERT(item->maybe_null());
     DBUG_ASSERT(item->null_value);
-    make_sort_key_longlong(to, item->maybe_null, true,
+    make_sort_key_longlong(to, item->maybe_null(), true,
                            item->unsigned_flag, 0);
   }
   else
-    make_sort_key_longlong(to, item->maybe_null, false,
+    make_sort_key_longlong(to, item->maybe_null(), false,
                            item->unsigned_flag, pack_time(&buf));
 }
 
@@ -1222,11 +1235,11 @@ Type_handler_timestamp_common::make_sort_key_part(uchar *to, Item *item,
   if (native.is_null() || native.is_zero_datetime())
   {
     // NULL or '0000-00-00 00:00:00'
-    bzero(to, item->maybe_null ? binlen + 1 : binlen);
+    bzero(to, item->maybe_null() ? binlen + 1 : binlen);
   }
   else
   {
-    if (item->maybe_null)
+    if (item->maybe_null())
       *to++= 1;
     if (native.length() != binlen)
     {
@@ -1310,7 +1323,7 @@ Type_handler_decimal_result::make_sort_key_part(uchar *to, Item *item,
                                             String *tmp_buffer) const
 {
   my_decimal dec_buf, *dec_val= item->val_decimal_result(&dec_buf);
-  if (item->maybe_null)
+  if (item->maybe_null())
   {
     if (item->null_value)
     {
@@ -1330,7 +1343,7 @@ Type_handler_real_result::make_sort_key_part(uchar *to, Item *item,
                                              String *tmp_buffer) const
 {
   double value= item->val_result();
-  if (item->maybe_null)
+  if (item->maybe_null())
   {
     if (item->null_value)
     {
@@ -2272,7 +2285,7 @@ sortlength(THD *thd, Sort_keys *sort_keys, bool *allow_packing_for_sortkeys)
                                             thd->variables.max_sort_length));
       }
 
-      if ((sortorder->maybe_null= sortorder->item->maybe_null))
+      if ((sortorder->maybe_null= sortorder->item->maybe_null()))
         nullable_cols++;				// Place for NULL marker
     }
     if (sortorder->is_variable_sized())
@@ -2576,7 +2589,7 @@ Type_handler_string_result::make_packed_sort_key_part(uchar *to, Item *item,
                                             String *tmp) const
 {
   CHARSET_INFO *cs= item->collation.collation;
-  bool maybe_null= item->maybe_null;
+  bool maybe_null= item->maybe_null();
 
   if (maybe_null)
     *to++= 1;
@@ -2615,7 +2628,7 @@ Type_handler_int_result::make_packed_sort_key_part(uchar *to, Item *item,
                                             String *tmp) const
 {
   longlong value= item->val_int_result();
-  return make_packed_sort_key_longlong(to, item->maybe_null,
+  return make_packed_sort_key_longlong(to, item->maybe_null(),
                                        item->null_value, item->unsigned_flag,
                                        value, sort_field);
 }
@@ -2627,7 +2640,7 @@ Type_handler_decimal_result::make_packed_sort_key_part(uchar *to, Item *item,
                                             String *tmp) const
 {
   my_decimal dec_buf, *dec_val= item->val_decimal_result(&dec_buf);
-  if (item->maybe_null)
+  if (item->maybe_null())
   {
     if (item->null_value)
     {
@@ -2649,7 +2662,7 @@ Type_handler_real_result::make_packed_sort_key_part(uchar *to, Item *item,
                                             String *tmp) const
 {
   double value= item->val_result();
-  if (item->maybe_null)
+  if (item->maybe_null())
   {
     if (item->null_value)
     {
@@ -2675,12 +2688,12 @@ Type_handler_temporal_result::make_packed_sort_key_part(uchar *to, Item *item,
   static const Temporal::Options opt(TIME_INVALID_DATES, TIME_FRAC_NONE);
   if (item->get_date_result(current_thd, &buf, opt))
   {
-    DBUG_ASSERT(item->maybe_null);
+    DBUG_ASSERT(item->maybe_null());
     DBUG_ASSERT(item->null_value);
-    return make_packed_sort_key_longlong(to, item->maybe_null, true,
+    return make_packed_sort_key_longlong(to, item->maybe_null(), true,
                                          item->unsigned_flag, 0, sort_field);
   }
-  return make_packed_sort_key_longlong(to, item->maybe_null, false,
+  return make_packed_sort_key_longlong(to, item->maybe_null(), false,
                                        item->unsigned_flag, pack_time(&buf),
                                        sort_field);
 }
@@ -2697,7 +2710,7 @@ Type_handler_timestamp_common::make_packed_sort_key_part(uchar *to, Item *item,
   if (native.is_null() || native.is_zero_datetime())
   {
     // NULL or '0000-00-00 00:00:00'
-    if (item->maybe_null)
+    if (item->maybe_null())
     {
       *to++=0;
       return 0;
@@ -2710,7 +2723,7 @@ Type_handler_timestamp_common::make_packed_sort_key_part(uchar *to, Item *item,
   }
   else
   {
-    if (item->maybe_null)
+    if (item->maybe_null())
       *to++= 1;
     if (native.length() != binlen)
     {
@@ -3035,7 +3048,7 @@ static uint make_sortkey(Sort_param *param, uchar *to)
                                                            sort_field->item,
                                                            sort_field,
                                                            &param->tmp_buffer);
-      if ((maybe_null= sort_field->item->maybe_null))
+      if ((maybe_null= sort_field->item->maybe_null()))
         to++;
     }
 
@@ -3088,7 +3101,7 @@ static uint make_packed_sortkey(Sort_param *param, uchar *to)
       length= item->type_handler()->make_packed_sort_key_part(to, item,
                                                               sort_field,
                                                               &param->tmp_buffer);
-      if ((maybe_null= sort_field->item->maybe_null))
+      if ((maybe_null= sort_field->item->maybe_null()))
         to++;
     }
     to+= length;
