@@ -46,6 +46,7 @@ Created 12/9/1995 Heikki Tuuri
 #include "buf0dump.h"
 #include "log0sync.h"
 #include "log.h"
+#include "tpool.h"
 
 /*
 General philosophy of InnoDB redo-logs:
@@ -69,9 +70,7 @@ log_t	log_sys;
 
 void log_t::set_capacity()
 {
-#ifndef SUX_LOCK_GENERIC
-	ut_ad(log_sys.latch.is_write_locked());
-#endif
+	ut_ad(log_sys.latch_have_wr());
 	/* Margin for the free space in the smallest log, before a new query
 	step which modifies the database, is started */
 
@@ -88,11 +87,7 @@ void log_t::set_capacity()
 	log_sys.max_checkpoint_age = margin;
 }
 
-#ifdef HAVE_PMEM
-void log_t::create_low()
-#else
-bool log_t::create()
-#endif
+void log_t::create()
 {
   ut_ad(this == &log_sys);
   ut_ad(!is_initialised());
@@ -103,38 +98,14 @@ bool log_t::create()
   need_checkpoint.store(true, std::memory_order_relaxed);
   write_lsn= FIRST_LSN;
 
-#ifndef HAVE_PMEM
-  buf= static_cast<byte*>(ut_malloc_dontdump(buf_size, PSI_INSTRUMENT_ME));
-  if (!buf)
-  {
-  alloc_fail:
-    sql_print_error("InnoDB: Cannot allocate memory;"
-                    " too large innodb_log_buffer_size?");
-    return false;
-  }
-  flush_buf= static_cast<byte*>(ut_malloc_dontdump(buf_size,
-                                                   PSI_INSTRUMENT_ME));
-  if (!flush_buf)
-  {
-    ut_free_dodump(buf, buf_size);
-    buf= nullptr;
-    goto alloc_fail;
-  }
-
-  TRASH_ALLOC(buf, buf_size);
-  TRASH_ALLOC(flush_buf, buf_size);
-  checkpoint_buf= static_cast<byte*>(aligned_malloc(4096, 4096));
-  memset_aligned<4096>(checkpoint_buf, 0, 4096);
-  max_buf_free= buf_size / LOG_BUF_FLUSH_RATIO - LOG_BUF_FLUSH_MARGIN;
-#else
   ut_ad(!checkpoint_buf);
   ut_ad(!buf);
   ut_ad(!flush_buf);
+  ut_ad(!writer);
   max_buf_free= 1;
-#endif
 
   latch.SRW_LOCK_INIT(log_latch_key);
-  init_lsn_lock();
+  lsn_lock.init();
 
   last_checkpoint_lsn= FIRST_LSN;
   log_capacity= 0;
@@ -143,12 +114,9 @@ bool log_t::create()
   next_checkpoint_lsn= 0;
   checkpoint_pending= false;
 
-  buf_free= 0;
+  set_buf_free(0);
 
   ut_ad(is_initialised());
-#ifndef HAVE_PMEM
-  return true;
-#endif
 }
 
 dberr_t log_file_t::close() noexcept
@@ -166,35 +134,153 @@ __attribute__((warn_unused_result))
 dberr_t log_file_t::read(os_offset_t offset, span<byte> buf) noexcept
 {
   ut_ad(is_opened());
-  return os_file_read(IORequestRead, m_file, buf.data(), offset, buf.size(),
-                      nullptr);
+  byte *data= buf.data();
+  size_t size= buf.size();
+  ut_ad(size);
+  ssize_t s;
+
+  for (;;)
+  {
+    s= IF_WIN(tpool::pread(m_file, data, size, offset),
+              pread(m_file, data, size, offset));
+    if (UNIV_UNLIKELY(s <= 0))
+      break;
+    size-= size_t(s);
+    if (!size)
+      return DB_SUCCESS;
+    offset+= s;
+    data+= s;
+    ut_a(size < buf.size());
+  }
+
+  sql_print_error("InnoDB: pread(\"ib_logfile0\") returned %zd,"
+                  " operating system error %u",
+                  s, unsigned(IF_WIN(GetLastError(), errno)));
+  return DB_IO_ERROR;
 }
 
 void log_file_t::write(os_offset_t offset, span<const byte> buf) noexcept
 {
   ut_ad(is_opened());
-  if (dberr_t err= os_file_write_func(IORequestWrite, "ib_logfile0", m_file,
-                                      buf.data(), offset, buf.size()))
-    ib::fatal() << "write(\"ib_logfile0\") returned " << err;
+  const byte *data= buf.data();
+  size_t size= buf.size();
+  ut_ad(size);
+  ssize_t s;
+
+  for (;;)
+  {
+    s= IF_WIN(tpool::pwrite(m_file, data, size, offset),
+              pwrite(m_file, data, size, offset));
+    if (UNIV_UNLIKELY(s <= 0))
+      break;
+    size-= size_t(s);
+    if (!size)
+      return;
+    offset+= s;
+    data+= s;
+    ut_a(size < buf.size());
+  }
+
+  sql_print_error("[FATAL] InnoDB: pwrite(\"ib_logfile0\") returned %zd,"
+                  " operating system error %u",
+                  s, unsigned(IF_WIN(GetLastError(), errno)));
+  abort();
 }
 
-#ifdef HAVE_PMEM
-# include <libpmem.h>
+# ifdef HAVE_PMEM
+#  include "cache.h"
+# endif
 
 /** Attempt to memory map a file.
 @param file  log file handle
 @param size  file size
 @return pointer to memory mapping
 @retval MAP_FAILED  if the memory cannot be mapped */
-static void *log_mmap(os_file_t file, os_offset_t size)
+static void *log_mmap(os_file_t file,
+# ifdef HAVE_PMEM
+                      bool &is_pmem, /*!< whether the file is on pmem */
+# endif
+                      os_offset_t size)
 {
-  void *ptr=
-    my_mmap(0, size_t(size),
-            srv_read_only_mode ? PROT_READ : PROT_READ | PROT_WRITE,
-            MAP_SHARED_VALIDATE | MAP_SYNC, file, 0);
-#ifdef __linux__
-  if (ptr == MAP_FAILED)
+#if SIZEOF_SIZE_T < 8
+  if (size != os_offset_t(size_t(size)))
+    return MAP_FAILED;
+#endif
+  if (my_system_page_size > 4096)
+    return MAP_FAILED;
+# ifndef HAVE_PMEM
+  if (!log_sys.log_mmap)
+    /* If support for persistent memory (Linux: mount -o dax) is enabled,
+    we always attempt to open a MAP_SYNC memory mapping to ib_logfile0.
+    This mapping will be read-only during crash recovery, and read-write
+    during normal operation.
+
+    A regular read-only memory mapping may be attempted if
+    innodb_log_file_mmap=ON. This may benefit mariadb-backup
+    and crash recovery. */
+    return MAP_FAILED;
+# endif
+
+  /* For now, InnoDB does not support memory-mapped writes to
+  a regular log file.
+
+  If PMEM is supported, the initially attempted memory mapping may
+  be read-write, but the fallback will be read-only.
+
+  The mapping will always be read-only if innodb_read_only=ON or
+  if mariadb-backup is running in any other mode than --prepare --export. */
+  const bool read_only=
+    srv_read_only_mode || srv_operation >= SRV_OPERATION_BACKUP;
+
+# ifdef _WIN32
+  void *ptr= MAP_FAILED;
+  if (!read_only);
+  else if (HANDLE h=
+           CreateFileMappingA(file, nullptr, PAGE_READONLY,
+                              DWORD(size >> 32), DWORD(size), nullptr))
   {
+    if (h != INVALID_HANDLE_VALUE)
+    {
+      ptr= MapViewOfFileEx(h, FILE_MAP_READ, 0, 0, size, nullptr);
+      CloseHandle(h);
+      if (!ptr)
+        ptr= MAP_FAILED;
+    }
+  }
+# else
+  int flags=
+#  ifdef HAVE_PMEM
+    MAP_SHARED_VALIDATE | MAP_SYNC,
+#  else
+    MAP_SHARED,
+#  endif
+    prot= PROT_READ;
+
+  if (!read_only)
+#  ifdef HAVE_PMEM
+    prot= PROT_READ | PROT_WRITE;
+
+#   ifdef __linux__ /* On Linux, we pretend that /dev/shm is PMEM */
+remap:
+#   endif
+#  else
+    return MAP_FAILED;
+#  endif
+
+  void *ptr= my_mmap(0, size_t(size), prot, flags, file, 0);
+
+#  ifdef HAVE_PMEM
+  is_pmem= ptr != MAP_FAILED;
+#  endif
+
+  if (ptr != MAP_FAILED)
+    return ptr;
+
+#  ifdef HAVE_PMEM
+#   ifdef __linux__ /* On Linux, we pretend that /dev/shm is PMEM */
+  if (flags != MAP_SHARED && srv_operation < SRV_OPERATION_BACKUP)
+  {
+    flags= MAP_SHARED;
     struct stat st;
     if (!fstat(file, &st))
     {
@@ -203,48 +289,79 @@ static void *log_mmap(os_file_t file, os_offset_t size)
       if (!stat("/dev/shm", &st))
       {
         MSAN_STAT_WORKAROUND(&st);
-        if (st.st_dev == st_dev)
-          ptr= my_mmap(0, size_t(size),
-                       srv_read_only_mode ? PROT_READ : PROT_READ | PROT_WRITE,
-                       MAP_SHARED, file, 0);
+        is_pmem= st.st_dev == st_dev;
+        if (!is_pmem)
+          return ptr; /* MAP_FAILED */
+        goto remap;
       }
     }
   }
-#endif /* __linux__ */
+#   endif /* __linux__ */
+  if (read_only && log_sys.log_mmap)
+    ptr= my_mmap(0, size_t(size), PROT_READ, MAP_SHARED, file, 0);
+#  endif /* HAVE_PMEM */
+# endif
   return ptr;
 }
+
+#if defined __linux__ || defined _WIN32
+/** Display a message about opening the log */
+ATTRIBUTE_COLD static void log_file_message()
+{
+  sql_print_information("InnoDB: %s (block size=%u bytes)",
+                        log_sys.log_mmap
+                        ? (log_sys.log_buffered
+                           ? "Memory-mapped log"
+                           : "Memory-mapped unbuffered log")
+                        :
+                        log_sys.log_buffered
+                        ? "Buffered log writes"
+                        : "File system buffers for log disabled",
+                        log_sys.write_size);
+}
+#else
+static inline void log_file_message() {}
 #endif
 
-#ifdef HAVE_PMEM
 bool log_t::attach(log_file_t file, os_offset_t size)
-#else
-void log_t::attach_low(log_file_t file, os_offset_t size)
-#endif
 {
   log= file;
   ut_ad(!size || size >= START_OFFSET + SIZE_OF_FILE_CHECKPOINT);
   file_size= size;
 
-#ifdef HAVE_PMEM
   ut_ad(!buf);
   ut_ad(!flush_buf);
-  if (size && !(size_t(size) & 4095) && srv_operation != SRV_OPERATION_BACKUP)
+  ut_ad(!writer);
+  if (size)
   {
-    void *ptr= log_mmap(log.m_file, size);
+# ifdef HAVE_PMEM
+    bool is_pmem;
+    void *ptr= ::log_mmap(log.m_file, is_pmem, size);
+# else
+    void *ptr= ::log_mmap(log.m_file, size);
+# endif
     if (ptr != MAP_FAILED)
     {
-      log.close();
-      mprotect(ptr, size_t(size), PROT_READ);
-      buf= static_cast<byte*>(ptr);
-      max_buf_free= size;
-# if defined __linux__ || defined _WIN32
-      set_block_size(CPU_LEVEL1_DCACHE_LINESIZE);
+# ifdef HAVE_PMEM
+      if (is_pmem)
+      {
+        log.close();
+        log_buffered= false;
+        log_maybe_unbuffered= true;
+        IF_WIN(,mprotect(ptr, size_t(size), PROT_READ));
+      }
 # endif
-      log_maybe_unbuffered= true;
-      log_buffered= false;
-      return true;
+      buf= static_cast<byte*>(ptr);
+      max_buf_free= 1;
+      writer_update();
+# ifdef HAVE_PMEM
+      if (is_pmem)
+        return true;
+# endif
+      goto func_exit;
     }
   }
+  log_mmap= false;
   buf= static_cast<byte*>(ut_malloc_dontdump(buf_size, PSI_INSTRUMENT_ME));
   if (!buf)
   {
@@ -258,29 +375,32 @@ void log_t::attach_low(log_file_t file, os_offset_t size)
                                                    PSI_INSTRUMENT_ME));
   if (!flush_buf)
   {
+  alloc_fail2:
     ut_free_dodump(buf, buf_size);
     buf= nullptr;
     goto alloc_fail;
   }
 
+  ut_ad(ut_is_2pow(write_size));
+  ut_ad(write_size >= 512);
+  ut_ad(write_size <= 4096);
+  checkpoint_buf= static_cast<byte*>(aligned_malloc(write_size, write_size));
+  if (!checkpoint_buf)
+  {
+    ut_free_dodump(flush_buf, buf_size);
+    flush_buf= nullptr;
+    goto alloc_fail2;
+  }
+
   TRASH_ALLOC(buf, buf_size);
   TRASH_ALLOC(flush_buf, buf_size);
   max_buf_free= buf_size / LOG_BUF_FLUSH_RATIO - LOG_BUF_FLUSH_MARGIN;
-#endif
+  writer_update();
+  memset_aligned<512>(checkpoint_buf, 0, write_size);
 
-#if defined __linux__ || defined _WIN32
-  sql_print_information("InnoDB: %s (block size=%u bytes)",
-                        log_buffered
-                        ? "Buffered log writes"
-                        : "File system buffers for log disabled",
-                        block_size);
-#endif
-
-#ifdef HAVE_PMEM
-  checkpoint_buf= static_cast<byte*>(aligned_malloc(block_size, block_size));
-  memset_aligned<64>(checkpoint_buf, 0, block_size);
+ func_exit:
+  log_file_message();
   return true;
-#endif
 }
 
 /** Write a log file header.
@@ -311,9 +431,7 @@ void log_t::header_write(byte *buf, lsn_t lsn, bool encrypted)
 
 void log_t::create(lsn_t lsn) noexcept
 {
-#ifndef SUX_LOCK_GENERIC
-  ut_ad(latch.is_write_locked());
-#endif
+  ut_ad(latch_have_wr());
   ut_ad(!recv_no_log_write);
   ut_ad(is_latest());
   ut_ad(this == &log_sys);
@@ -325,66 +443,77 @@ void log_t::create(lsn_t lsn) noexcept
 
   last_checkpoint_lsn= 0;
 
-#ifdef HAVE_PMEM
-  if (is_pmem())
-  {
-    mprotect(buf, size_t(file_size), PROT_READ | PROT_WRITE);
-    memset_aligned<4096>(buf, 0, 4096);
-    buf_free= START_OFFSET;
-  }
-  else
-#endif
-  {
-    buf_free= 0;
-    memset_aligned<4096>(flush_buf, 0, buf_size);
-    memset_aligned<4096>(buf, 0, buf_size);
-  }
-
-  log_sys.header_write(buf, lsn, is_encrypted());
   DBUG_PRINT("ib_log", ("write header " LSN_PF, lsn));
 
 #ifdef HAVE_PMEM
-  if (is_pmem())
+  if (is_mmap())
+  {
+    ut_ad(!is_opened());
+    mprotect(buf, size_t(file_size), PROT_READ | PROT_WRITE);
+    memset_aligned<4096>(buf, 0, 4096);
+    log_sys.header_write(buf, lsn, is_encrypted());
+    set_buf_free(START_OFFSET);
     pmem_persist(buf, 512);
+  }
   else
 #endif
   {
+    ut_ad(!is_mmap());
+    set_buf_free(0);
+    memset_aligned<4096>(flush_buf, 0, buf_size);
+    memset_aligned<4096>(buf, 0, buf_size);
+    log_sys.header_write(buf, lsn, is_encrypted());
     log.write(0, {buf, 4096});
     memset_aligned<512>(buf, 0, 512);
   }
 }
 
-void log_t::close_file()
+ATTRIBUTE_COLD static void log_close_failed(dberr_t err)
 {
-#ifdef HAVE_PMEM
-  if (is_pmem())
+  ib::fatal() << "closing ib_logfile0 failed: " << err;
+}
+
+void log_t::close_file(bool really_close)
+{
+  if (is_mmap())
   {
-    ut_ad(!is_opened());
     ut_ad(!checkpoint_buf);
+    ut_ad(!flush_buf);
     if (buf)
     {
-      my_munmap(buf, file_size);
+      my_munmap(buf, size_t(file_size));
       buf= nullptr;
     }
-    return;
+  }
+  else
+  {
+    ut_ad(!buf == !flush_buf);
+    ut_ad(!buf == !checkpoint_buf);
+    if (buf)
+    {
+      ut_free_dodump(buf, buf_size);
+      buf= nullptr;
+      ut_free_dodump(flush_buf, buf_size);
+      flush_buf= nullptr;
+    }
+    aligned_free(checkpoint_buf);
+    checkpoint_buf= nullptr;
   }
 
-  ut_free_dodump(buf, buf_size);
-  buf= nullptr;
-  ut_free_dodump(flush_buf, buf_size);
-  flush_buf= nullptr;
-  aligned_free(checkpoint_buf);
-  checkpoint_buf= nullptr;
-#endif
-  if (is_opened())
-    if (const dberr_t err= log.close())
-      ib::fatal() << "closing ib_logfile0 failed: " << err;
+  writer= nullptr;
+
+  if (really_close)
+    if (is_opened())
+      if (const dberr_t err= log.close())
+        log_close_failed(err);
 }
 
 /** Acquire all latches that protect the log. */
 static void log_resize_acquire()
 {
-  if (!log_sys.is_pmem())
+#ifdef HAVE_PMEM
+  if (!log_sys.is_mmap())
+#endif
   {
     while (flush_lock.acquire(log_sys.get_lsn() + 1, nullptr) !=
            group_commit_lock::ACQUIRED);
@@ -400,7 +529,9 @@ void log_resize_release()
 {
   log_sys.latch.wr_unlock();
 
-  if (!log_sys.is_pmem())
+#ifdef HAVE_PMEM
+  if (!log_sys.is_mmap())
+#endif
   {
     lsn_t lsn1= write_lock.release(write_lock.value());
     lsn_t lsn2= flush_lock.release(flush_lock.value());
@@ -413,25 +544,25 @@ void log_resize_release()
 /** Try to enable or disable file system caching (update log_buffered) */
 void log_t::set_buffered(bool buffered)
 {
-  if (!log_maybe_unbuffered || is_pmem() || high_level_read_only)
+  if (!log_maybe_unbuffered ||
+#ifdef HAVE_PMEM
+      is_mmap() ||
+#endif
+      high_level_read_only)
     return;
   log_resize_acquire();
   if (!resize_in_progress() && is_opened() && bool(log_buffered) != buffered)
   {
-    os_file_close_func(log.m_file);
-    log.m_file= OS_FILE_CLOSED;
+    if (const dberr_t err= log.close())
+      log_close_failed(err);
     std::string path{get_log_file_path()};
     log_buffered= buffered;
     bool success;
     log.m_file= os_file_create_func(path.c_str(),
-                                    OS_FILE_OPEN, OS_FILE_NORMAL, OS_LOG_FILE,
+                                    OS_FILE_OPEN, OS_LOG_FILE,
                                     false, &success);
     ut_a(log.m_file != OS_FILE_CLOSED);
-    sql_print_information("InnoDB: %s (block size=%u bytes)",
-                          log_buffered
-                          ? "Buffered log writes"
-                          : "File system buffers for log disabled",
-                          block_size);
+    log_file_message();
   }
   log_resize_release();
 }
@@ -450,6 +581,9 @@ log_t::resize_start_status log_t::resize_start(os_offset_t size) noexcept
 
   resize_start_status status= RESIZE_NO_CHANGE;
   lsn_t start_lsn{0};
+#ifdef HAVE_PMEM
+  bool is_pmem{false};
+#endif
 
   if (resize_in_progress())
     status= RESIZE_IN_PROGRESS;
@@ -464,26 +598,29 @@ log_t::resize_start_status log_t::resize_start(os_offset_t size) noexcept
     resize_lsn.store(1, std::memory_order_relaxed);
     resize_target= 0;
     resize_log.m_file=
-      os_file_create_func(path.c_str(),
-                          OS_FILE_CREATE | OS_FILE_ON_ERROR_NO_EXIT,
-                          OS_FILE_NORMAL, OS_LOG_FILE, false, &success);
+      os_file_create_func(path.c_str(), OS_FILE_CREATE,
+                          OS_LOG_FILE, false, &success);
     if (success)
     {
+      ut_ad(!(size_t(file_size) & (write_size - 1)));
+      ut_ad(!(size_t(size) & (write_size - 1)));
       log_resize_release();
 
       void *ptr= nullptr, *ptr2= nullptr;
       success= os_file_set_size(path.c_str(), resize_log.m_file, size);
       if (!success);
 #ifdef HAVE_PMEM
-      else if (is_pmem())
+      else if (is_mmap())
       {
-        ptr= log_mmap(resize_log.m_file, size);
+        ptr= ::log_mmap(resize_log.m_file, is_pmem, size);
+
         if (ptr == MAP_FAILED)
           goto alloc_fail;
       }
 #endif
       else
       {
+        ut_ad(!is_mmap());
         ptr= ut_malloc_dontdump(buf_size, PSI_INSTRUMENT_ME);
         if (ptr)
         {
@@ -515,21 +652,20 @@ log_t::resize_start_status log_t::resize_start(os_offset_t size) noexcept
         resize_target= size;
         resize_buf= static_cast<byte*>(ptr);
         resize_flush_buf= static_cast<byte*>(ptr2);
-        if (is_pmem())
-        {
-          resize_log.close();
-          start_lsn= get_lsn();
-        }
-        else
-        {
-          memcpy_aligned<16>(resize_buf, buf, (buf_free + 15) & ~15);
+        start_lsn= get_lsn();
+
+        if (!is_mmap())
           start_lsn= first_lsn +
-            (~lsn_t{get_block_size() - 1} & (write_lsn - first_lsn));
-        }
+            (~lsn_t{write_size - 1} &
+             (lsn_t{write_size - 1} + start_lsn - first_lsn));
+        else if (!is_opened())
+          resize_log.close();
+
+        writer_update();
       }
-      resize_lsn.store(start_lsn, std::memory_order_relaxed);
       status= success ? RESIZE_STARTED : RESIZE_FAILED;
     }
+    resize_lsn.store(start_lsn, std::memory_order_relaxed);
   }
 
   log_resize_release();
@@ -554,58 +690,65 @@ void log_t::resize_abort() noexcept
 
   if (resize_in_progress() > 1)
   {
-    if (!is_pmem())
+#ifdef HAVE_PMEM
+    const bool is_mmap{this->is_mmap()};
+#else
+    constexpr bool is_mmap{false};
+#endif
+    if (!is_mmap)
     {
-      resize_log.close();
       ut_free_dodump(resize_buf, buf_size);
       ut_free_dodump(resize_flush_buf, buf_size);
       resize_flush_buf= nullptr;
     }
-#ifdef HAVE_PMEM
     else
     {
       ut_ad(!resize_log.is_opened());
       ut_ad(!resize_flush_buf);
+#ifdef HAVE_PMEM
       if (resize_buf)
         my_munmap(resize_buf, resize_target);
+#endif /* HAVE_PMEM */
     }
-#endif
+    if (resize_log.is_opened())
+      resize_log.close();
     resize_buf= nullptr;
     resize_target= 0;
     resize_lsn.store(0, std::memory_order_relaxed);
+    std::string path{get_log_file_path("ib_logfile101")};
+    IF_WIN(DeleteFile(path.c_str()), unlink(path.c_str()));
   }
 
+  writer_update();
   log_resize_release();
 }
 
 /** Write an aligned buffer to ib_logfile0.
 @param buf    buffer to be written
-@param len    length of data to be written
+@param length length of data to be written
 @param offset log file offset */
-static void log_write_buf(const byte *buf, size_t len, lsn_t offset)
+static void log_write_buf(const byte *buf, size_t length, lsn_t offset)
 {
   ut_ad(write_lock.is_owner());
   ut_ad(!recv_no_log_write);
-  ut_d(const size_t block_size_1= log_sys.get_block_size() - 1);
+  ut_d(const size_t block_size_1= log_sys.write_size - 1);
   ut_ad(!(offset & block_size_1));
-  ut_ad(!(len & block_size_1));
+  ut_ad(!(length & block_size_1));
   ut_ad(!(size_t(buf) & block_size_1));
-  ut_ad(len);
+  ut_ad(length);
 
-  if (UNIV_LIKELY(offset + len <= log_sys.file_size))
+  const lsn_t maximum_write_length{log_sys.file_size - offset};
+  ut_ad(maximum_write_length <= log_sys.file_size - log_sys.START_OFFSET);
+
+  if (UNIV_UNLIKELY(length > maximum_write_length))
   {
-write:
-    log_sys.log.write(offset, {buf, len});
-    return;
+    log_sys.log.write(offset, {buf, size_t(maximum_write_length)});
+    length-= size_t(maximum_write_length);
+    buf+= size_t(maximum_write_length);
+    ut_ad(log_sys.START_OFFSET + length < offset);
+    offset= log_sys.START_OFFSET;
   }
-
-  const size_t write_len= size_t(log_sys.file_size - offset);
-  log_sys.log.write(offset, {buf, write_len});
-  len-= write_len;
-  buf+= write_len;
-  ut_ad(log_sys.START_OFFSET + len < offset);
-  offset= log_sys.START_OFFSET;
-  goto write;
+  log_sys.log.write(offset, {buf, length});
 }
 
 /** Invoke commit_checkpoint_notify_ha() to notify that outstanding
@@ -734,33 +877,30 @@ static size_t log_pad(lsn_t lsn, size_t pad, byte *begin, byte *extra)
 #endif
 
 #ifdef HAVE_PMEM
-/** Persist the log.
-@param lsn    desired new value of flushed_to_disk_lsn */
-inline void log_t::persist(lsn_t lsn) noexcept
+void log_t::persist(lsn_t lsn) noexcept
 {
-  ut_ad(is_pmem());
+  ut_ad(!is_opened());
   ut_ad(!write_lock.is_owner());
   ut_ad(!flush_lock.is_owner());
+#ifdef LOG_LATCH_DEBUG
+  ut_ad(latch_have_any());
+#endif
 
   lsn_t old= flushed_to_disk_lsn.load(std::memory_order_relaxed);
 
   if (old >= lsn)
     return;
 
-  const lsn_t resizing{resize_in_progress()};
-  if (UNIV_UNLIKELY(resizing))
-    latch.rd_lock(SRW_LOCK_CALL);
   const size_t start(calc_lsn_offset(old));
   const size_t end(calc_lsn_offset(lsn));
 
   if (UNIV_UNLIKELY(end < start))
   {
-    pmem_persist(log_sys.buf + start, log_sys.file_size - start);
-    pmem_persist(log_sys.buf + log_sys.START_OFFSET,
-                 end - log_sys.START_OFFSET);
+    pmem_persist(buf + start, file_size - start);
+    pmem_persist(buf + START_OFFSET, end - START_OFFSET);
   }
   else
-    pmem_persist(log_sys.buf + start, end - start);
+    pmem_persist(buf + start, end - start);
 
   old= flushed_to_disk_lsn.load(std::memory_order_relaxed);
 
@@ -774,25 +914,37 @@ inline void log_t::persist(lsn_t lsn) noexcept
     log_flush_notify(lsn);
     DBUG_EXECUTE_IF("crash_after_log_write_upto", DBUG_SUICIDE(););
   }
+}
 
-  if (UNIV_UNLIKELY(resizing))
-    latch.rd_unlock();
+ATTRIBUTE_NOINLINE
+static void log_write_persist(lsn_t lsn) noexcept
+{
+  log_sys.latch.rd_lock(SRW_LOCK_CALL);
+  log_sys.persist(lsn);
+  log_sys.latch.rd_unlock();
 }
 #endif
 
-/** Write resize_buf to resize_log.
-@param length  the used length of resize_buf */
-ATTRIBUTE_COLD void log_t::resize_write_buf(size_t length) noexcept
+void log_t::resize_write_buf(const byte *b, size_t length) noexcept
 {
-  const size_t block_size_1= get_block_size() - 1;
+  const size_t block_size_1= write_size - 1;
+  ut_ad(b == resize_buf || b == resize_flush_buf);
   ut_ad(!(resize_target & block_size_1));
   ut_ad(!(length & block_size_1));
   ut_ad(length > block_size_1);
   ut_ad(length <= resize_target);
-  const lsn_t resizing{resize_in_progress()};
-  ut_ad(resizing <= write_lsn);
-  lsn_t offset= START_OFFSET +
-    ((write_lsn - resizing) & ~lsn_t{block_size_1}) %
+
+  int64_t d= int64_t(write_lsn - resize_in_progress());
+  if (UNIV_UNLIKELY(d < 0))
+  {
+    d&= ~int64_t(block_size_1);
+    if (int64_t(d + length) <= 0)
+      return;
+    length+= ssize_t(d);
+    b-= ssize_t(d);
+    d= 0;
+  }
+  lsn_t offset= START_OFFSET + (lsn_t(d) & ~lsn_t{block_size_1}) %
     (resize_target - START_OFFSET);
 
   if (UNIV_UNLIKELY(offset + length > resize_target))
@@ -804,88 +956,126 @@ ATTRIBUTE_COLD void log_t::resize_write_buf(size_t length) noexcept
   }
 
   ut_a(os_file_write_func(IORequestWrite, "ib_logfile101", resize_log.m_file,
-                          resize_flush_buf, offset, length) == DB_SUCCESS);
+                          b, offset, length) == DB_SUCCESS);
 }
 
-/** Write buf to ib_logfile0.
-@tparam release_latch whether to invoke latch.wr_unlock()
+/** Write buf to ib_logfile0 and possibly ib_logfile101.
+@tparam resizing whether to release latch and whether resize_in_progress()>1
 @return the current log sequence number */
-template<bool release_latch> inline lsn_t log_t::write_buf() noexcept
+template<log_t::resizing_and_latch resizing>
+inline __attribute__((always_inline))
+lsn_t log_t::write_buf() noexcept
 {
-#ifndef SUX_LOCK_GENERIC
-  ut_ad(latch.is_write_locked());
-#endif
-  ut_ad(!is_pmem());
+  ut_ad(latch_have_wr());
+  ut_ad(!is_mmap());
   ut_ad(!srv_read_only_mode);
+  ut_ad(resizing == RETAIN_LATCH ||
+        (resizing == RESIZING) == (resize_in_progress() > 1));
 
   const lsn_t lsn{get_lsn(std::memory_order_relaxed)};
 
   if (write_lsn >= lsn)
   {
-    if (release_latch)
+    if (resizing != RETAIN_LATCH)
       latch.wr_unlock();
     ut_ad(write_lsn == lsn);
   }
   else
   {
+    ut_ad(write_lock.is_owner());
     ut_ad(!recv_no_log_write);
     write_lock.set_pending(lsn);
     ut_ad(write_lsn >= get_flushed_lsn());
-    const size_t block_size_1{get_block_size() - 1};
-    lsn_t offset{calc_lsn_offset(write_lsn) & ~lsn_t{block_size_1}};
+    const size_t write_size_1{write_size - 1};
+    ut_ad(ut_is_2pow(write_size));
+    size_t length{buf_free.load(std::memory_order_relaxed)};
+    lsn_t offset{calc_lsn_offset(write_lsn)};
+    ut_ad(length >= (offset & write_size_1));
+    ut_ad(write_size_1 >= 511);
+
+    const byte *const write_buf{buf};
+    byte *const re_write_buf{resizing == NOT_RESIZING ? nullptr : resize_buf};
+    ut_ad(resizing == RETAIN_LATCH ||
+          (resizing == NOT_RESIZING) == !re_write_buf);
+    ut_ad(!re_write_buf == !resize_flush_buf);
+    if (resizing == RESIZING)
+#ifdef _MSC_VER
+      __assume(re_write_buf != nullptr);
+#else
+      if (!re_write_buf) __builtin_unreachable();
+#endif
+    offset&= ~lsn_t{write_size_1};
+
+    if (length <= write_size_1)
+    {
+      ut_ad(!((length ^ (size_t(lsn) - size_t(first_lsn))) & write_size_1));
+      /* Keep filling the same buffer until we have more than one block. */
+#if 0 /* TODO: Pad the last log block with dummy records. */
+      buf_free= log_pad(lsn, (write_size_1 + 1) - length,
+                        buf + length, flush_buf);
+      ... /* TODO: Update the LSN and adjust other code. */
+#else
+      MEM_MAKE_DEFINED(buf + length, (write_size_1 + 1) - length);
+      buf[length]= 0; /* ensure that recovery catches EOF */
+#endif
+      if (UNIV_LIKELY_NULL(re_write_buf))
+      {
+        MEM_MAKE_DEFINED(re_write_buf + length, (write_size_1 + 1) - length);
+        re_write_buf[length]= 0;
+      }
+      length= write_size_1 + 1;
+    }
+    else
+    {
+      const size_t new_buf_free{length & write_size_1};
+      ut_ad(new_buf_free == ((lsn - first_lsn) & write_size_1));
+      buf_free.store(new_buf_free, std::memory_order_relaxed);
+
+      if (new_buf_free)
+      {
+        /* The rest of the block will be written as garbage.
+        (We want to avoid memset() while holding exclusive log_sys.latch)
+        This block will be overwritten later, once records beyond
+        the current LSN are generated. */
+        MEM_MAKE_DEFINED(buf + length, (write_size_1 + 1) - new_buf_free);
+        buf[length]= 0; /* allow recovery to catch EOF faster */
+        length&= ~write_size_1;
+        memcpy_aligned<16>(flush_buf, buf + length, (new_buf_free + 15) & ~15);
+        if (UNIV_LIKELY_NULL(re_write_buf))
+        {
+          MEM_MAKE_DEFINED(re_write_buf + length, (write_size_1 + 1) -
+                           new_buf_free);
+          memcpy_aligned<16>(resize_flush_buf, re_write_buf + length,
+                             (new_buf_free + 15) & ~15);
+          re_write_buf[length + new_buf_free]= 0;
+        }
+        length+= write_size_1 + 1;
+      }
+
+      std::swap(buf, flush_buf);
+      if (UNIV_LIKELY_NULL(re_write_buf))
+        std::swap(resize_buf, resize_flush_buf);
+    }
+
+    write_to_log++;
+    if (resizing != RETAIN_LATCH)
+      latch.wr_unlock();
 
     DBUG_PRINT("ib_log", ("write " LSN_PF " to " LSN_PF " at " LSN_PF,
                           write_lsn, lsn, offset));
-    const byte *write_buf{buf};
-    size_t length{buf_free};
-    ut_ad(length >= (calc_lsn_offset(write_lsn) & block_size_1));
-    const size_t new_buf_free{length & block_size_1};
-    buf_free= new_buf_free;
-    ut_ad(new_buf_free == ((lsn - first_lsn) & block_size_1));
 
-    if (new_buf_free)
-    {
-#if 0 /* TODO: Pad the last log block with dummy records. */
-      buf_free= log_pad(lsn, get_block_size() - new_buf_free,
-                        buf + new_buf_free, flush_buf);
-      ... /* TODO: Update the LSN and adjust other code. */
-#else
-      /* The rest of the block will be written as garbage.
-      (We want to avoid memset() while holding exclusive log_sys.latch)
-      This block will be overwritten later, once records beyond
-      the current LSN are generated. */
-# ifdef HAVE_valgrind
-      MEM_MAKE_DEFINED(buf + length, get_block_size() - new_buf_free);
-      if (UNIV_LIKELY_NULL(resize_flush_buf))
-        MEM_MAKE_DEFINED(resize_buf + length, get_block_size() - new_buf_free);
-# endif
-      buf[length]= 0; /* allow recovery to catch EOF faster */
-      length&= ~block_size_1;
-      memcpy_aligned<16>(flush_buf, buf + length, (new_buf_free + 15) & ~15);
-      if (UNIV_LIKELY_NULL(resize_flush_buf))
-        memcpy_aligned<16>(resize_flush_buf, resize_buf + length,
-                           (new_buf_free + 15) & ~15);
-      length+= get_block_size();
-#endif
-    }
+    /* Do the write to the log file */
+    log_write_buf(write_buf, length, offset);
 
-    std::swap(buf, flush_buf);
-    std::swap(resize_buf, resize_flush_buf);
-    write_to_log++;
-    if (release_latch)
-      latch.wr_unlock();
+    if (UNIV_LIKELY_NULL(re_write_buf))
+      resize_write_buf(re_write_buf, length);
+    write_lsn= lsn;
 
     if (UNIV_UNLIKELY(srv_shutdown_state > SRV_SHUTDOWN_INITIATED))
     {
       service_manager_extend_timeout(INNODB_EXTEND_TIMEOUT_INTERVAL,
                                      "InnoDB log write: " LSN_PF, write_lsn);
     }
-
-    /* Do the write to the log file */
-    log_write_buf(write_buf, length, offset);
-    if (UNIV_LIKELY_NULL(resize_buf))
-      resize_write_buf(length);
-    write_lsn= lsn;
   }
 
   set_check_for_checkpoint(false);
@@ -913,7 +1103,7 @@ bool log_t::flush(lsn_t lsn) noexcept
 */
 static lsn_t log_flush(lsn_t lsn)
 {
-  ut_ad(!log_sys.is_pmem());
+  ut_ad(!log_sys.is_mmap());
   ut_a(log_sys.flush(lsn));
   DBUG_EXECUTE_IF("crash_after_log_write_upto", DBUG_SUICIDE(););
   return flush_lock.release(lsn);
@@ -928,11 +1118,12 @@ wait and check if an already running write is covering the request.
 @param durable  whether the write needs to be durable
 @param callback log write completion callback */
 void log_write_up_to(lsn_t lsn, bool durable,
-                     const completion_callback *callback)
+                     const completion_callback *callback) noexcept
 {
-  ut_ad(!srv_read_only_mode || (log_sys.buf_free < log_sys.max_buf_free));
+  ut_ad(!srv_read_only_mode || log_sys.buf_free_ok());
   ut_ad(lsn != LSN_MAX);
   ut_ad(lsn != 0);
+  ut_ad(!log_sys.is_mmap() || !callback || durable);
 
   if (UNIV_UNLIKELY(recv_no_ibuf_operations))
   {
@@ -945,21 +1136,25 @@ void log_write_up_to(lsn_t lsn, bool durable,
   ut_ad(lsn <= log_sys.get_lsn());
 
 #ifdef HAVE_PMEM
-  if (log_sys.is_pmem())
+  if (log_sys.is_mmap())
   {
-    ut_ad(!callback);
     if (durable)
-      log_sys.persist(lsn);
+      log_write_persist(lsn);
     return;
   }
 #endif
+  ut_ad(!log_sys.is_mmap());
 
 repeat:
   if (durable)
   {
     if (flush_lock.acquire(lsn, callback) != group_commit_lock::ACQUIRED)
       return;
-    flush_lock.set_pending(log_sys.get_lsn());
+    /* Promise to other concurrent flush_lock.acquire() that we
+    will durable at least up to the current LSN. The LSN may still
+    advance until we acquire log_sys.latch below. */
+    lsn= log_sys.get_lsn();
+    flush_lock.set_pending(lsn);
   }
 
   lsn_t pending_write_lsn= 0, pending_flush_lsn= 0;
@@ -968,7 +1163,7 @@ repeat:
       group_commit_lock::ACQUIRED)
   {
     log_sys.latch.wr_lock(SRW_LOCK_CALL);
-    pending_write_lsn= write_lock.release(log_sys.write_buf<true>());
+    pending_write_lsn= write_lock.release(log_sys.writer());
   }
 
   if (durable)
@@ -985,6 +1180,23 @@ repeat:
   }
 }
 
+static lsn_t log_writer() noexcept
+{
+  return log_sys.write_buf<log_t::NOT_RESIZING>();
+}
+
+ATTRIBUTE_COLD static lsn_t log_writer_resizing() noexcept
+{
+  return log_sys.write_buf<log_t::RESIZING>();
+}
+
+void log_t::writer_update() noexcept
+{
+  ut_ad(latch_have_wr());
+  writer= resize_in_progress() ? log_writer_resizing : log_writer;
+  mtr_t::finisher_update();
+}
+
 /** Write to the log file up to the last log entry.
 @param durable  whether to wait for a durable write to complete */
 void log_buffer_flush_to_disk(bool durable)
@@ -995,8 +1207,10 @@ void log_buffer_flush_to_disk(bool durable)
 /** Prepare to invoke log_write_and_flush(), before acquiring log_sys.latch. */
 ATTRIBUTE_COLD void log_write_and_flush_prepare()
 {
-  if (log_sys.is_pmem())
+#ifdef HAVE_PMEM
+  if (log_sys.is_mmap())
     return;
+#endif
 
   while (flush_lock.acquire(log_sys.get_lsn() + 1, nullptr) !=
          group_commit_lock::ACQUIRED);
@@ -1004,20 +1218,54 @@ ATTRIBUTE_COLD void log_write_and_flush_prepare()
          group_commit_lock::ACQUIRED);
 }
 
+void log_t::clear_mmap()
+{
+  if (!is_mmap() ||
+#ifdef HAVE_PMEM
+      !is_opened() ||
+#endif
+      high_level_read_only)
+    return;
+  log_resize_acquire();
+  ut_ad(!resize_in_progress());
+  ut_ad(write_lsn == get_lsn());
+  ut_ad(write_lsn == get_flushed_lsn(std::memory_order_relaxed));
+
+  if (buf) /* this may be invoked while creating a new database */
+  {
+    alignas(16) byte log_block[4096];
+    const size_t bs{write_size};
+    const size_t bf{buf_free.load(std::memory_order_relaxed)};
+    {
+      byte *const b= buf;
+      memcpy_aligned<16>(log_block, b + (bf & ~(bs - 1)), bs);
+    }
+
+    close_file(false);
+    log_mmap= false;
+    ut_a(attach(log, file_size));
+    ut_ad(!is_mmap());
+
+    set_buf_free(bf & (bs - 1));
+    memcpy_aligned<16>(log_sys.buf, log_block, bs);
+  }
+  log_resize_release();
+}
+
 /** Durably write the log up to log_sys.get_lsn(). */
 ATTRIBUTE_COLD void log_write_and_flush()
 {
   ut_ad(!srv_read_only_mode);
-  if (!log_sys.is_pmem())
+#ifdef HAVE_PMEM
+  if (log_sys.is_mmap())
+    log_sys.persist(log_sys.get_lsn());
+  else
+#endif
   {
-    const lsn_t lsn{log_sys.write_buf<false>()};
+    const lsn_t lsn{log_sys.write_buf<log_t::RETAIN_LATCH>()};
     write_lock.release(lsn);
     log_flush(lsn);
   }
-#ifdef HAVE_PMEM
-  else
-    log_sys.persist(log_sys.get_lsn());
-#endif
 }
 
 /****************************************************************//**
@@ -1067,7 +1315,7 @@ NOTE that this function may only be called while not holding
 any synchronization objects except dict_sys.latch. */
 void log_free_check()
 {
-  ut_ad(!lock_sys.is_writer());
+  ut_ad(!lock_sys.is_holder());
   if (log_sys.check_for_checkpoint())
   {
     ut_ad(!recv_no_log_write);
@@ -1291,24 +1539,16 @@ log_print(
 void log_t::close()
 {
   ut_ad(this == &log_sys);
+  ut_ad(!(buf_free & buf_free_LOCK));
   if (!is_initialised()) return;
   close_file();
 
-#ifndef HAVE_PMEM
-  ut_free_dodump(buf, buf_size);
-  buf= nullptr;
-  ut_free_dodump(flush_buf, buf_size);
-  flush_buf= nullptr;
-  aligned_free(checkpoint_buf);
-  checkpoint_buf= nullptr;
-#else
   ut_ad(!checkpoint_buf);
   ut_ad(!buf);
   ut_ad(!flush_buf);
-#endif
 
   latch.destroy();
-  destroy_lsn_lock();
+  lsn_lock.destroy();
 
   recv_sys.close();
 
