@@ -17,7 +17,6 @@
 #define MYSQL_SERVER 1
 #include <my_global.h>
 #include "mysql_version.h"
-#include "spd_environ.h"
 #include "sql_priv.h"
 #include "probes_mysql.h"
 #include "sql_class.h"
@@ -218,25 +217,25 @@ int spider_udf_direct_sql_create_conn_key(
         tables_on_different_db_are_joinable();
     direct_sql->conn_key_length
       = 1
-      + direct_sql->tgt_wrapper_length + 1
-      + direct_sql->tgt_host_length + 1
-      + 5 + 1
-      + direct_sql->tgt_socket_length + 1
-      + (tables_on_different_db_are_joinable ?
-        0 : direct_sql->tgt_default_db_name_length + 1)
-      + direct_sql->tgt_username_length + 1
-      + direct_sql->tgt_password_length + 1
-      + direct_sql->tgt_ssl_ca_length + 1
-      + direct_sql->tgt_ssl_capath_length + 1
-      + direct_sql->tgt_ssl_cert_length + 1
-      + direct_sql->tgt_ssl_cipher_length + 1
-      + direct_sql->tgt_ssl_key_length + 1
-      + 1 + 1
-      + direct_sql->tgt_default_file_length + 1
-      + direct_sql->tgt_default_group_length + 1
-      + direct_sql->tgt_dsn_length + 1
-      + direct_sql->tgt_filedsn_length + 1
-      + direct_sql->tgt_driver_length;
+      + (direct_sql->tgt_wrapper ? direct_sql->tgt_wrapper_length + 2 : 0)
+      + (direct_sql->tgt_host ? direct_sql->tgt_host_length + 2 : 0)
+      + 5 + 2
+      + (direct_sql->tgt_socket ? direct_sql->tgt_socket_length + 2 : 0)
+      + (!tables_on_different_db_are_joinable && direct_sql->tgt_default_db_name ?
+         direct_sql->tgt_default_db_name_length + 2 : 0)
+      + (direct_sql->tgt_username ? direct_sql->tgt_username_length + 2 : 0)
+      + (direct_sql->tgt_password ? direct_sql->tgt_password_length + 2 : 0)
+      + (direct_sql->tgt_ssl_ca ? direct_sql->tgt_ssl_ca_length + 2 : 0)
+      + (direct_sql->tgt_ssl_capath ? direct_sql->tgt_ssl_capath_length + 2 : 0)
+      + (direct_sql->tgt_ssl_cert ? direct_sql->tgt_ssl_cert_length + 2 : 0)
+      + (direct_sql->tgt_ssl_cipher ? direct_sql->tgt_ssl_cipher_length + 2 : 0)
+      + (direct_sql->tgt_ssl_key ? direct_sql->tgt_ssl_key_length + 2 : 0)
+      + 1 + 2
+      + (direct_sql->tgt_default_file ? direct_sql->tgt_default_file_length + 2 : 0)
+      + (direct_sql->tgt_default_group ? direct_sql->tgt_default_group_length + 2 : 0)
+      + (direct_sql->tgt_dsn ? direct_sql->tgt_dsn_length + 2 : 0)
+      + (direct_sql->tgt_filedsn ? direct_sql->tgt_filedsn_length + 2 : 0)
+      + (direct_sql->tgt_driver ? direct_sql->tgt_driver_length + 2 : 0);
   if (!(direct_sql->conn_key = (char *)
     spider_malloc(spider_current_trx, SPD_MID_UDF_DIRECT_SQL_CREATE_CONN_KEY_1, direct_sql->conn_key_length + 1,
       MYF(MY_WME | MY_ZEROFILL)))
@@ -246,124 +245,60 @@ int spider_udf_direct_sql_create_conn_key(
     *direct_sql->conn_key = '0' + 48 - direct_sql->connection_channel;
   else
     *direct_sql->conn_key = '0' + direct_sql->connection_channel;
-  DBUG_PRINT("info",("spider tgt_wrapper=%s", direct_sql->tgt_wrapper));
-  tmp_name = strmov(direct_sql->conn_key + 1, direct_sql->tgt_wrapper);
-  if (direct_sql->tgt_host)
-  {
-    DBUG_PRINT("info",("spider tgt_host=%s", direct_sql->tgt_host));
-    tmp_name = strmov(tmp_name + 1, direct_sql->tgt_host);
-  } else {
-    DBUG_PRINT("info",("spider tgt_host=NULL"));
-    tmp_name++;
-  }
+  int counter= 0;
+  tmp_name= direct_sql->conn_key + 1;
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_wrapper);
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_host);
   my_sprintf(port_str, (port_str, "%05ld", direct_sql->tgt_port));
-  DBUG_PRINT("info",("spider port_str=%s", port_str));
-  tmp_name = strmov(tmp_name + 1, port_str);
-  if (direct_sql->tgt_socket)
-  {
-    DBUG_PRINT("info",("spider tgt_socket=%s", direct_sql->tgt_socket));
-    tmp_name = strmov(tmp_name + 1, direct_sql->tgt_socket);
-  }
-  else
+  spider_create_conn_key_add_one(&counter, &tmp_name, port_str);
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_socket);
+  counter++;
+  if (!tables_on_different_db_are_joinable && direct_sql->tgt_default_db_name)
   {
+    *tmp_name= (char) counter;
+    tmp_name = strmov(tmp_name + 1, direct_sql->tgt_default_db_name);
     tmp_name++;
   }
-    if (!tables_on_different_db_are_joinable)
-    {
-      if (direct_sql->tgt_default_db_name)
-      {
-        DBUG_PRINT("info",("spider tgt_default_db_name=%s",
-          direct_sql->tgt_default_db_name));
-        tmp_name = strmov(tmp_name + 1, direct_sql->tgt_default_db_name);
-      } else
-        tmp_name++;
-    }
-    if (direct_sql->tgt_username)
-    {
-      DBUG_PRINT("info",("spider tgt_username=%s", direct_sql->tgt_username));
-      tmp_name = strmov(tmp_name + 1, direct_sql->tgt_username);
-    } else
-      tmp_name++;
-    if (direct_sql->tgt_password)
-    {
-      DBUG_PRINT("info",("spider tgt_password=%s", direct_sql->tgt_password));
-      tmp_name = strmov(tmp_name + 1, direct_sql->tgt_password);
-    } else
-      tmp_name++;
-    if (direct_sql->tgt_ssl_ca)
-    {
-      DBUG_PRINT("info",("spider tgt_ssl_ca=%s", direct_sql->tgt_ssl_ca));
-      tmp_name = strmov(tmp_name + 1, direct_sql->tgt_ssl_ca);
-    } else
-      tmp_name++;
-    if (direct_sql->tgt_ssl_capath)
-    {
-      DBUG_PRINT("info",("spider tgt_ssl_capath=%s",
-        direct_sql->tgt_ssl_capath));
-      tmp_name = strmov(tmp_name + 1, direct_sql->tgt_ssl_capath);
-    } else
-      tmp_name++;
-    if (direct_sql->tgt_ssl_cert)
-    {
-      DBUG_PRINT("info",("spider tgt_ssl_cert=%s", direct_sql->tgt_ssl_cert));
-      tmp_name = strmov(tmp_name + 1, direct_sql->tgt_ssl_cert);
-    } else
-      tmp_name++;
-    if (direct_sql->tgt_ssl_cipher)
-    {
-      DBUG_PRINT("info",("spider tgt_ssl_cipher=%s",
-        direct_sql->tgt_ssl_cipher));
-      tmp_name = strmov(tmp_name + 1, direct_sql->tgt_ssl_cipher);
-    } else
-      tmp_name++;
-    if (direct_sql->tgt_ssl_key)
-    {
-      DBUG_PRINT("info",("spider tgt_ssl_key=%s", direct_sql->tgt_ssl_key));
-      tmp_name = strmov(tmp_name + 1, direct_sql->tgt_ssl_key);
-    } else
-      tmp_name++;
-    tmp_name++;
-    *tmp_name = '0' + ((char) direct_sql->tgt_ssl_vsc);
-    if (direct_sql->tgt_default_file)
-    {
-      DBUG_PRINT("info",("spider tgt_default_file=%s",
-        direct_sql->tgt_default_file));
-      tmp_name = strmov(tmp_name + 1, direct_sql->tgt_default_file);
-    } else
-      tmp_name++;
-    if (direct_sql->tgt_default_group)
-    {
-      DBUG_PRINT("info",("spider tgt_default_group=%s",
-        direct_sql->tgt_default_group));
-      tmp_name = strmov(tmp_name + 1, direct_sql->tgt_default_group);
-    } else
-      tmp_name++;
-    if (direct_sql->tgt_dsn)
-    {
-      DBUG_PRINT("info",("spider tgt_dsn=%s",
-        direct_sql->tgt_dsn));
-      tmp_name = strmov(tmp_name + 1, direct_sql->tgt_dsn);
-    } else
-      tmp_name++;
-    if (direct_sql->tgt_filedsn)
-    {
-      DBUG_PRINT("info",("spider tgt_filedsn=%s",
-        direct_sql->tgt_filedsn));
-      tmp_name = strmov(tmp_name + 1, direct_sql->tgt_filedsn);
-    } else
-      tmp_name++;
-    if (direct_sql->tgt_driver)
-    {
-      DBUG_PRINT("info",("spider tgt_driver=%s",
-        direct_sql->tgt_driver));
-      tmp_name = strmov(tmp_name + 1, direct_sql->tgt_driver);
-    } else
-      tmp_name++;
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_username);
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_password);
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_ssl_ca);
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_ssl_capath);
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_ssl_cert);
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_ssl_cipher);
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_ssl_key);
+  counter++;
+  *tmp_name= (char) counter;
+  tmp_name++;
+  *tmp_name = '0' + ((char) direct_sql->tgt_ssl_vsc);
+  tmp_name++;
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_default_file);
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_default_group);
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_dsn);
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_filedsn);
+  spider_create_conn_key_add_one(&counter, &tmp_name, direct_sql->tgt_driver);
+  tmp_name++;
   direct_sql->conn_key_hash_value = my_calc_hash(&spider_open_connections,
     (uchar*) direct_sql->conn_key, direct_sql->conn_key_length);
   DBUG_RETURN(0);
 }
 
+static inline void spider_maybe_memcpy_string(
+  char **dest,
+  char *src,
+  char *tmp,
+  uint *dest_len,
+  uint src_len)
+{
+    *dest_len= src_len;
+    if (src)
+    {
+      *dest= tmp;
+      memcpy(*dest, src, src_len);
+    } else
+      *dest= NULL;
+}
+
+
 SPIDER_CONN *spider_udf_direct_sql_create_conn(
   const SPIDER_DIRECT_SQL *direct_sql,
   int *error_num
@@ -425,113 +360,56 @@ SPIDER_CONN *spider_udf_direct_sql_create_conn(
   conn->conn_key_length = direct_sql->conn_key_length;
   conn->conn_key = tmp_name;
   memcpy(conn->conn_key, direct_sql->conn_key, direct_sql->conn_key_length);
-  conn->tgt_wrapper_length = direct_sql->tgt_wrapper_length;
-  conn->tgt_wrapper = tmp_wrapper;
-  memcpy(conn->tgt_wrapper, direct_sql->tgt_wrapper,
-    direct_sql->tgt_wrapper_length);
-  conn->tgt_host_length = direct_sql->tgt_host_length;
-  conn->tgt_host = tmp_host;
-  memcpy(conn->tgt_host, direct_sql->tgt_host, direct_sql->tgt_host_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_wrapper, direct_sql->tgt_wrapper, tmp_wrapper,
+      &conn->tgt_wrapper_length, direct_sql->tgt_wrapper_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_host, direct_sql->tgt_host, tmp_host,
+      &conn->tgt_host_length, direct_sql->tgt_host_length);
     conn->tgt_port = direct_sql->tgt_port;
-    conn->tgt_socket_length = direct_sql->tgt_socket_length;
-    conn->tgt_socket = tmp_socket;
-    memcpy(conn->tgt_socket, direct_sql->tgt_socket,
-      direct_sql->tgt_socket_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_socket, direct_sql->tgt_socket, tmp_socket,
+      &conn->tgt_socket_length, direct_sql->tgt_socket_length);
     if (!tables_on_different_db_are_joinable)
-    {
-      conn->tgt_db_length = direct_sql->tgt_default_db_name_length;
-      conn->tgt_db = tmp_db;
-      memcpy(conn->tgt_db, direct_sql->tgt_default_db_name,
-        direct_sql->tgt_default_db_name_length);
-    }
-    conn->tgt_username_length = direct_sql->tgt_username_length;
-    conn->tgt_username = tmp_username;
-    memcpy(conn->tgt_username, direct_sql->tgt_username,
-      direct_sql->tgt_username_length);
-    conn->tgt_password_length = direct_sql->tgt_password_length;
-    conn->tgt_password = tmp_password;
-    memcpy(conn->tgt_password, direct_sql->tgt_password,
-      direct_sql->tgt_password_length);
-    conn->tgt_ssl_ca_length = direct_sql->tgt_ssl_ca_length;
-    if (conn->tgt_ssl_ca_length)
-    {
-      conn->tgt_ssl_ca = tmp_ssl_ca;
-      memcpy(conn->tgt_ssl_ca, direct_sql->tgt_ssl_ca,
-        direct_sql->tgt_ssl_ca_length);
-    } else
-      conn->tgt_ssl_ca = NULL;
-    conn->tgt_ssl_capath_length = direct_sql->tgt_ssl_capath_length;
-    if (conn->tgt_ssl_capath_length)
-    {
-      conn->tgt_ssl_capath = tmp_ssl_capath;
-      memcpy(conn->tgt_ssl_capath, direct_sql->tgt_ssl_capath,
-        direct_sql->tgt_ssl_capath_length);
-    } else
-      conn->tgt_ssl_capath = NULL;
-    conn->tgt_ssl_cert_length = direct_sql->tgt_ssl_cert_length;
-    if (conn->tgt_ssl_cert_length)
-    {
-      conn->tgt_ssl_cert = tmp_ssl_cert;
-      memcpy(conn->tgt_ssl_cert, direct_sql->tgt_ssl_cert,
-        direct_sql->tgt_ssl_cert_length);
-    } else
-      conn->tgt_ssl_cert = NULL;
-    conn->tgt_ssl_cipher_length = direct_sql->tgt_ssl_cipher_length;
-    if (conn->tgt_ssl_cipher_length)
-    {
-      conn->tgt_ssl_cipher = tmp_ssl_cipher;
-      memcpy(conn->tgt_ssl_cipher, direct_sql->tgt_ssl_cipher,
-        direct_sql->tgt_ssl_cipher_length);
-    } else
-      conn->tgt_ssl_cipher = NULL;
-    conn->tgt_ssl_key_length = direct_sql->tgt_ssl_key_length;
-    if (conn->tgt_ssl_key_length)
-    {
-      conn->tgt_ssl_key = tmp_ssl_key;
-      memcpy(conn->tgt_ssl_key, direct_sql->tgt_ssl_key,
-        direct_sql->tgt_ssl_key_length);
-    } else
-      conn->tgt_ssl_key = NULL;
-    conn->tgt_default_file_length = direct_sql->tgt_default_file_length;
-    if (conn->tgt_default_file_length)
-    {
-      conn->tgt_default_file = tmp_default_file;
-      memcpy(conn->tgt_default_file, direct_sql->tgt_default_file,
-        direct_sql->tgt_default_file_length);
-    } else
-      conn->tgt_default_file = NULL;
-    conn->tgt_default_group_length = direct_sql->tgt_default_group_length;
-    if (conn->tgt_default_group_length)
-    {
-      conn->tgt_default_group = tmp_default_group;
-      memcpy(conn->tgt_default_group, direct_sql->tgt_default_group,
-        direct_sql->tgt_default_group_length);
-    } else
-      conn->tgt_default_group = NULL;
-    conn->tgt_dsn_length = direct_sql->tgt_dsn_length;
-    if (conn->tgt_dsn_length)
-    {
-      conn->tgt_dsn = tmp_dsn;
-      memcpy(conn->tgt_dsn, direct_sql->tgt_dsn,
-        direct_sql->tgt_dsn_length);
-    } else
-      conn->tgt_dsn = NULL;
-    conn->tgt_filedsn_length = direct_sql->tgt_filedsn_length;
-    if (conn->tgt_filedsn_length)
-    {
-      conn->tgt_filedsn = tmp_filedsn;
-      memcpy(conn->tgt_filedsn, direct_sql->tgt_filedsn,
-        direct_sql->tgt_filedsn_length);
-    } else
-      conn->tgt_filedsn = NULL;
-    conn->tgt_driver_length = direct_sql->tgt_driver_length;
-    if (conn->tgt_driver_length)
-    {
-      conn->tgt_driver = tmp_driver;
-      memcpy(conn->tgt_driver, direct_sql->tgt_driver,
-        direct_sql->tgt_driver_length);
-    } else
-      conn->tgt_driver = NULL;
+      spider_maybe_memcpy_string(
+        &conn->tgt_db, direct_sql->tgt_default_db_name, tmp_db,
+        &conn->tgt_db_length, direct_sql->tgt_default_db_name_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_username, direct_sql->tgt_username, tmp_username,
+      &conn->tgt_username_length, direct_sql->tgt_username_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_password, direct_sql->tgt_password, tmp_password,
+      &conn->tgt_password_length, direct_sql->tgt_password_length);
+     spider_maybe_memcpy_string(
+      &conn->tgt_ssl_ca, direct_sql->tgt_ssl_ca, tmp_ssl_ca,
+      &conn->tgt_ssl_ca_length, direct_sql->tgt_ssl_ca_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_ssl_capath, direct_sql->tgt_ssl_capath, tmp_ssl_capath,
+      &conn->tgt_ssl_capath_length, direct_sql->tgt_ssl_capath_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_ssl_cert, direct_sql->tgt_ssl_cert, tmp_ssl_cert,
+      &conn->tgt_ssl_cert_length, direct_sql->tgt_ssl_cert_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_ssl_cipher, direct_sql->tgt_ssl_cipher, tmp_ssl_cipher,
+      &conn->tgt_ssl_cipher_length, direct_sql->tgt_ssl_cipher_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_ssl_key, direct_sql->tgt_ssl_key, tmp_ssl_key,
+      &conn->tgt_ssl_key_length, direct_sql->tgt_ssl_key_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_default_file, direct_sql->tgt_default_file, tmp_default_file,
+      &conn->tgt_default_file_length, direct_sql->tgt_default_file_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_default_group, direct_sql->tgt_default_group, tmp_default_group,
+      &conn->tgt_default_group_length, direct_sql->tgt_default_group_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_dsn, direct_sql->tgt_dsn, tmp_dsn,
+      &conn->tgt_dsn_length, direct_sql->tgt_dsn_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_filedsn, direct_sql->tgt_filedsn, tmp_filedsn,
+      &conn->tgt_filedsn_length, direct_sql->tgt_filedsn_length);
+    spider_maybe_memcpy_string(
+      &conn->tgt_driver, direct_sql->tgt_driver, tmp_driver,
+      &conn->tgt_driver_length, direct_sql->tgt_driver_length);
     conn->tgt_ssl_vsc = direct_sql->tgt_ssl_vsc;
   conn->dbton_id = direct_sql->dbton_id;
   conn->conn_need_mon = need_mon;
@@ -551,7 +429,6 @@ SPIDER_CONN *spider_udf_direct_sql_create_conn(
   conn->semi_trx_isolation = -2;
   conn->semi_trx_isolation_chk = FALSE;
   conn->semi_trx_chk = FALSE;
-    conn->conn_kind = SPIDER_CONN_KIND_MYSQL;
 
   if (mysql_mutex_init(spd_key_mutex_mta_conn, &conn->mta_conn_mutex,
     MY_MUTEX_INIT_FAST))
@@ -612,7 +489,6 @@ SPIDER_CONN *spider_udf_direct_sql_create_conn(
   DBUG_RETURN(conn);
 
 error:
-  DBUG_ASSERT(!conn->mta_conn_mutex_file_pos.file_name);
 error_too_many_ipport_count:
   spider_conn_done(conn);
 error_conn_init:
@@ -697,7 +573,6 @@ SPIDER_CONN *spider_udf_direct_sql_get_conn(
     conn->queued_ping = FALSE;
 
   DBUG_PRINT("info",("spider conn=%p", conn));
-  DBUG_PRINT("info",("spider conn->conn_kind=%u", conn->conn_kind));
   DBUG_RETURN(conn);
 
 error:
@@ -1206,7 +1081,7 @@ int spider_udf_set_direct_sql_param_default(
     direct_sql->tgt_default_db_name_length = SPIDER_THD_db_length(trx->thd);
     if (
       !(direct_sql->tgt_default_db_name = spider_create_string(
-        SPIDER_THD_db_str(trx->thd),
+          trx->thd->get_db(),
         direct_sql->tgt_default_db_name_length))
     ) {
       my_error(ER_OUT_OF_RESOURCES, MYF(0), HA_ERR_OUT_OF_MEM);
@@ -1422,8 +1297,8 @@ void spider_udf_free_direct_sql_alloc(
 long long spider_direct_sql_body(
   UDF_INIT *initid,
   UDF_ARGS *args,
-  char *is_null,
-  char *error,
+  unsigned char *is_null,
+  unsigned char *error,
   my_bool bg
 ) {
   int error_num, roop_count;
@@ -1658,6 +1533,11 @@ my_bool spider_direct_sql_init_body(
 ) {
   SPIDER_BG_DIRECT_SQL *bg_direct_sql;
   DBUG_ENTER("spider_direct_sql_init_body");
+  if (!spider_hton_ptr)
+  {
+    strcpy(message, "Plugin 'SPIDER' is not loaded");
+    goto error;
+  }
   if (args->arg_count != 3)
   {
     strcpy(message, "spider_(bg)_direct_sql() requires 3 arguments");
