@@ -211,7 +211,8 @@ static int
 my_strnncollsp_nchars_simple(CHARSET_INFO * cs,
                              const uchar *a, size_t a_length,
                              const uchar *b, size_t b_length,
-                             size_t nchars)
+                             size_t nchars,
+                             uint flags)
 {
   set_if_smaller(a_length, nchars);
   set_if_smaller(b_length, nchars);
@@ -756,7 +757,10 @@ ulonglong my_strntoull_8bit(CHARSET_INFO *cs,
     return (~(ulonglong) 0);
   }
 
-  return (negative ? -((longlong) i) : (longlong) i);
+  /* Avoid undefinite behavior - negation of LONGLONG_MIN */
+  return negative && (longlong) i != LONGLONG_MIN ?
+        -((longlong) i) :
+         (longlong) i;
 
 noconv:
   err[0]= EDOM;
@@ -1483,8 +1487,6 @@ static my_bool
 my_cset_init_8bit(struct charset_info_st *cs, MY_CHARSET_LOADER *loader)
 {
   cs->state|= my_8bit_charset_flags_from_data(cs);
-  cs->caseup_multiply= 1;
-  cs->casedn_multiply= 1;
   cs->pad_char= ' ';
   if (!cs->to_lower || !cs->to_upper || !cs->m_ctype || !cs->tab_to_uni)
     return TRUE;
@@ -2144,7 +2146,9 @@ MY_CHARSET_HANDLER my_charset_8bit_handler=
     my_well_formed_char_length_8bit,
     my_copy_8bit,
     my_wc_mb_bin, /* native_to_mb */
-    my_wc_to_printable_8bit
+    my_wc_to_printable_8bit,
+    my_casefold_multiply_1,
+    my_casefold_multiply_1
 };
 
 MY_COLLATION_HANDLER my_collation_8bit_simple_ci_handler =
