@@ -617,7 +617,10 @@ bs:
     return (~(ulonglong) 0);
   }
 
-  return (negative ? -((longlong) res) : (longlong) res);
+  /* Avoid undefinite behavior - negation of LONGLONG_MIN */
+  return negative && (longlong) res != LONGLONG_MIN ?
+         -((longlong) res) :
+          (longlong) res;
 }
 
 
@@ -1009,6 +1012,8 @@ end4:
   {
    if (li > MAX_NEGATIVE_NUMBER)
      goto overflow;
+   if (li == MAX_NEGATIVE_NUMBER) // Avoid undefinite behavior in negation
+     return LONGLONG_MIN;
    return -((longlong) li);
   }
   return (longlong) li;
@@ -2513,6 +2518,8 @@ end4:
   {
    if (li > MAX_NEGATIVE_NUMBER)
      goto overflow;
+   if (li == MAX_NEGATIVE_NUMBER) // Avoid undefinite behavior in negation
+     return LONGLONG_MIN;
    return -((longlong) li);
   }
   return (longlong) li;
